/* libgoth/vector.c
 *
 * A general linear array structure.
 *
 * Copyright (C) 2003, 2004
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

/* === Header files. ====================================================== */

#include <stdlib.h>

#include "libgoth/error.h"
#include "libgoth/misc.h"
#include "libgoth/vector.h"

/* === Non-static function definitions. =================================== */

/* Creates a new empty vector with room for 'size' elements, each 'sizeof_elem'
 * bytes in length.  Its address is stored at *'b'. */
int vector_new(vector_t** v, int size, int sizeof_elem)
{
    vector_t* vn;
    if (v == NULL || size < 0 || sizeof_elem <= 0) return -E_INVAL;

    vn = malloc(sizeof(*vn));
    if (vn == NULL) return -E_MEMORY;

    if (size > 0) {
        vn->data = malloc(size * sizeof_elem);
        if (vn->data == NULL) {free(vn); return -E_MEMORY;}
    } else {
        vn->data = NULL;
    }

    vn->count = 0;
    vn->capacity = size;
    vn->sizeof_elem = sizeof_elem;
    vn->sort_func = NULL;

    *v = vn;
    return E_OK;
}

/* Deallocates vector 'v'. */
int vector_del(vector_t* v)
{
    if (v == NULL) return -E_INVAL;
    if (v->data != NULL) free(v->data);
    free(v);
    return E_OK;
}

/* Resizes vector 'v' to have room for 'new_size' elements.  Returns 'new_size'
 * on success and an error code on failure. */
int vector_resize(vector_t* v, int new_size)
{
    if (v == NULL || new_size < v->count) return -E_INVAL;

    if (new_size != v->capacity) {
        if (new_size == 0) {
            /* Zero capacity. */
            free(v->data);
            v->data = NULL;
        } else {
            /* Nonzero capacity. */
            char* new_data = realloc(v->data, new_size * v->sizeof_elem);
            if (new_data == NULL) return -E_MEMORY;
            v->data = new_data;
        }
    }

    v->capacity = new_size;
    return new_size;
}

/* Adds 'data', a pointer to an array of 'count' elements, to the vector 'v'
 * at position 'pos'.  Negative positions are relative to the end of the
 * vector.  The vector is resized to hold all incoming data.  Returns the
 * number of elements written on success and an error code on failure. */
int vector_insert(vector_t* v, int pos, void* data, int count)
{
    if (v == NULL || data == NULL || count < 0) return -E_INVAL;
    else if (count == 0)                        return 0;

    /* Overallocation policy: get MAX(count * 3 / 2, 8) additional elements. */

    /* Normalize pos. */
    if (pos > v->count)       pos = v->count;
    else if (pos < -v->count) pos = 0;
    else if (pos < 0)         pos += v->count + 1;

    if (v->count + count > v->capacity) {
        /* Grow vector. */
        int ret = vector_resize(v, v->count + (count > 5 ? count * 3 / 2 : 8));
        if (ret < E_OK) return ret;
    }

    if (pos < v->count) {
        /* Move old data to make room for the new. */
        memmove(v->data + (pos + count) * v->sizeof_elem,
                v->data + pos * v->sizeof_elem,
                (v->count - pos) * v->sizeof_elem);
    }

    /* Copy the new data into place. */
    memcpy(v->data + pos * v->sizeof_elem, data, count * v->sizeof_elem);

    v->count += count;
    return count;
}

/* Copies 'count' elements from vector 'v' to the 'data' array.  Copied data is
 * removed from 'v'.  Returns the number of elements written on success and an
 * error code on failure.  Data is copied from position 'pos'.  If 'pos' is
 * negative, it is relative to the end of the vector. */
int vector_pop(vector_t* v, int pos, void* data, int count)
{
    int ret = vector_peek(v, pos, data, count);
    if (ret <= 0) return ret;
    else          return vector_discard(v, pos, count); 
}

/* Copies 'count' elements from vector 'v' to the 'data' array.  Returns the
 * number of elements written on success and an error code on failure.  Data
 * is copied from position 'pos'.  If 'pos' is negative, it is relative to
 * the end of the vector. */
int vector_peek(vector_t* v, int pos, void* data, int count)
{
    if (v == NULL || data == NULL || count < 0) return -E_INVAL;
    else if (count == 0)                        return 0;
    else if (count > v->count)                  count = v->count;

    /* Normalize pos. */
    if (pos > v->count - count)       pos = v->count - count;
    else if (pos < -v->count + count) pos = 0;
    else if (pos < 0)                 pos += v->count - count + 1;

    /* Copy. */
    memcpy(data, v->data + pos * v->sizeof_elem, count * v->sizeof_elem);

    return count;
}

/* Sets *'data' to point to a linear array of 'count' elements taken from the
 * vector 'v' starting at position 'pos'.  Returns the number of elements in
 * the array (may be more than 'count') on success and an error code on
 * failure.  If 'pos' is negative, it is relative to the end of the vector. */
int vector_index(vector_t* v, int pos, void* data, int count)
{
    if (v == NULL || data == NULL || count < 0) return -E_INVAL;
    else if (count > v->count)                  count = v->count;

    /* Normalize pos. */
    if (pos > v->count - count)       pos = v->count - count;
    else if (pos < -v->count + count) pos = 0;
    else if (pos < 0)                 pos += v->count - count + 1;

    *(void**)data = v->data + pos * v->sizeof_elem;

    return v->count - pos;
}

/* Throws away up to 'count' elements from vector 'v' at position 'pos'.
 * Returns the number of discarded elements on success and an error code on
 * failure.  If 'pos' is negative, it is relative to the end of the vector. */
int vector_discard(vector_t* v, int pos, int count)
{
    if (v == NULL || count < 0) return -E_INVAL;
    else if (count == 0)        return 0;
    else if (count > v->count)  count = v->count;

    /* Normalize pos. */
    if (pos > v->count - count)       pos = v->count - count;
    else if (pos < -v->count + count) pos = 0;
    else if (pos < 0)                 pos += v->count - count + 1;

    if (pos + count < v->count) {
        /* Overwrite data to be removed. */
        memmove(v->data + pos * v->sizeof_elem,
                v->data + (pos + count) * v->sizeof_elem,
                (v->count - pos - count) * v->sizeof_elem);
    }

    v->count -= count;
    return count;
}

/* Returns the number of elements currently stored in vector 'v'. */
int vector_count(vector_t* v)
{
    if (v == NULL) return -E_INVAL;
    else           return v->count;
}

/* Returns the current size in elements of vector 'v'. */
int vector_size(vector_t* v)
{
    if (v == NULL) return -E_INVAL;
    else           return v->capacity;
}

/* Sets vector 'v''s sort function to 'sort_func'. */
int vector_set_sort_func(vector_t* v, vector_sort_func_t sort_func)
{
    if (v == NULL) return -E_INVAL;
    v->sort_func = sort_func;
    return E_OK;
}

/* Sorts vector 'v'. */
int vector_sort(vector_t* v)
{
    if (v == NULL) return -E_INVAL;

    if (v->sort_func != NULL) {
        /* Hooray for libc. */
        qsort(v->data, v->count, v->sizeof_elem, v->sort_func);
    }

    return E_OK;
}

/* Searches for an element 'data' in vector 'v'.  Stores the index of the first
 * occurrence in *'first' and one plus the index of the last occurrence in
 * *'last'.  If 'data' is not found, 'first' equals 'last' and both point to
 * the position in which 'data' would be inserted to maintain sort order.  On
 * success, returns the number of occurrences of 'data' (which is equal to
 * *'last' minus *'first'), and on failure returns a negative error code.
 * 'first' and/or 'last' can be NULL, but this causes vector_find_sorted()
 * to return 0 on item-not-found and 1 on item-found rather than the total
 * number of occurrences. */
int vector_find_sorted(vector_t* v, void* data, int* first, int* last)
{
    int ret;
    int left, pos, right;
    int _first, _last;
    char* ptr;
    if (v == NULL || data == NULL || v->sort_func == NULL) return -E_INVAL;

    /* Find the first occurrence. */
    if (first != NULL || (first == NULL && last == NULL)) {
        left = 0; right = v->count;
        while (1) {
            pos = (left + right) / 2;
            if (left >= right) break;

            ret = v->sort_func(v->data + pos * v->sizeof_elem, data);
            if      (ret < 0) left  = pos + 1;  /* v[pos] < data */
            else if (ret > 0) right = pos;      /* v[pos] > data */
            else              left  = right;    /* v[pos] = data */
        }
        _first = pos;
    } else {
        pos = -1;
    }

    /* Find the last occurrence. */
    if (last != NULL) {
        left = pos + 1; right = v->count;
        while (1) {
            pos = (left + right) / 2;
            if (left >= right) break;

            ret = v->sort_func(v->data + pos * v->sizeof_elem, data);
            if      (ret < 0) left  = right;   /* v[pos] < data */
            else if (ret > 0) left  = pos + 1; /* v[pos] = data */
            else              right = pos;     /* v[pos] > data */
        }
        _last = pos;
    }

    /* Done! */
    if (first != NULL)      *first = _first;
    if (last  != NULL)      *last  = _last ;
    if (last == NULL)       ptr = v->data + _first * v->sizeof_elem;
    else if (first == NULL) ptr = v->data + (_last - 1) * v->sizeof_elem;
    else                    return _last - _first;

    /* Didn't request both first and last. */
    if (v->count == 0) return 0;
    else               return (v->sort_func(ptr, data) == 0) ? 1 : 0;
}

/* Inserts an element 'data' into vector 'v'.  The insertion position is
 * determined by sort_func().  Returns the insertion position on success and
 * a negative error code on failure. */
int vector_insert_sorted(vector_t* v, void* data)
{
    int ret, pos;
    if (v == NULL || data == NULL) return -E_INVAL;

    try vector_find_sorted(v, data, &pos, NULL); catch(error1);
    try vector_insert(v, pos, data, 1); catch(error1);

    return pos;
error1: return ret;
}

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
