/* dev/rfid.c
 *
 * RFID tag reader.
 *
 * Copyright (C) 2003, 2004
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

/* TODO:
 *
 * > Implement the full RFID protocol.  The main things that are missing:
 *
 *   - Tag querying.
 *   - Tag modification.
 *   - Zero-tags-found.
 *   - Anti-collision mode.
 *   - Checksum calculation and verification.
 *   - Error/status checking.
 *
 *   Most of these get turned into their own functions callable from the
 *   higher-level code.  In particular, tag querying will be handled by the
 *   timer callback that this code should register.  Oh, that reminds me, there
 *   needs to be a function or such to set the query frequency.
 *
 * > This code needs to be cleaned up. */

/* === Header files. ====================================================== */

#include <stdio.h>
#include <stdlib.h>

#include "dev/rfid.h"
#include "lib/device.h"
#include "lib/error.h"
#include "lib/misc.h"
#include "lib/vector.h"

/* === Type definitions, static variables, and static prototypes. ========= */

/* An individual tag read by the device. */
typedef struct tag {
    long long id;                   /* RFID.                                */
    int magic_cookie;               /* Used for traversal.                  */
} tag_t;

/* State information for an RFID tag reader device. */
typedef struct rfid {
    vector_t* tags;                 /* List of tags.  <tag_t>               */

    int packet_rem;                 /* Bytes remaining until end of packet. */

#if 0
    shelf_t* shelf;                 /* Associated shelf.                    */
#endif
} rfid_t;

static int rfid_on_packet(device_t* d, char* data, int len);
static int rfid_is_delim(device_t* d, char data);
static int rfid_on_cleanup(device_t* d);
static vector_sort_func_t rfid_sort_tag;

/* === Non-static function definitions. =================================== */

/* Connects 'd' to an RFID tag reader. */
int rfid_open(device_t* d)
{
    int ret, i;
    rfid_t* r;
    if (d == NULL) return -E_INVAL;

    r = malloc(sizeof(*r));
    if (r == NULL) return -E_MEMORY;

    /* Connect to the RFID-specific handlers. */
    d->on_packet      = rfid_on_packet;
    d->is_delim       = rfid_is_delim;
    d->on_cleanup_dev = rfid_on_cleanup;
    d->dev_state      = r;

    /* Initial settings. */
    r->packet_rem = 0;
    vector_new(&r->tags, 0, sizeof(tag_t));
    vector_set_sort_func(r->tags, rfid_sort_tag);

    return E_OK;
}

/* === Static function definitions. ======================================== */

static int rfid_on_packet(device_t* d, char* data, int len)
{
    int i, ret;
    int num_tags;
    int magic_cookie;
    int pos;
    tag_t* tags;
    tag_t tag;

    rfid_t* r = d->dev_state;
    /* I assume all packets just tell me tag information. */

    num_tags = vector_count(r->tags);
    if (num_tags > 0) {
        vector_index(r->tags, 0, &tags, num_tags);
        magic_cookie = tags[0].magic_cookie;
    } else {
        magic_cookie = 0;
    }

    num_tags = data[5];

    /* Check for added and removed tags. */
    for (i = 0; i < num_tags; ++i) {
        tag.id = ((long long)data[ 8 + 10 * i] << 48) |
                 ((long long)data[ 9 + 10 * i] << 40) |
                 ((long long)data[10 + 10 * i] << 32) |
                 ((long long)data[12 + 10 * i] << 24) |
                 ((long long)data[13 + 10 * i] << 16) |
                 ((long long)data[14 + 10 * i] <<  8) |
                 ((long long)data[15 + 10 * i] <<  0);
        tag.magic_cookie = !magic_cookie;

        ret = vector_find_sorted(r->tags, &tag, &pos, NULL);
        if (ret < E_OK) return ret;
        if (ret == 0) {
            /* This is a newly-added tag. */
            vector_insert_sorted(r->tags, &tag);
            vector_index(r->tags, 0, &tags, 0);
            printf("NEW %016llx\n", tag.id);
        } else {
            /* This tag was already on the shelf. */
            tags[pos].magic_cookie = !magic_cookie;
        }
    }

    /* Now remove all tags not visited by the above. */
    num_tags = vector_count(r->tags);
    for (i = 0; i < num_tags; ++i) {
        if (tags[i].magic_cookie == magic_cookie) {
            /* This tag was removed from the shelf. */
            vector_discard(r->tags, i, 1);
            printf("REM %016llx\n", tags[i].id);

            /* Update pointers. */
            --i; --num_tags;
            vector_index(r->tags, 0, &tags, 0);
        }
    }

    return E_OK;
}

static int rfid_is_delim(device_t* d, char data)
{
    rfid_t* r = d->com_state;
    if (r->packet_rem == 0) r->packet_rem = data;
    return (--r->packet_rem == 0) ? 1 : 0;
}

static int rfid_on_cleanup(device_t* d)
{
    rfid_t* r = d->dev_state;
    vector_del(r->tags);
    free(r);
    return E_OK;
}

static int _rfid_sort_tag(const tag_t* a, const tag_t* b)
{
    return SGN(a->id - b->id);
}
static vector_sort_func_t rfid_sort_tag = (vector_sort_func_t)_rfid_sort_tag;

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
