/* lib/device.c
 *
 * Device management and input/output abstraction.
 *
 * Copyright (C) 2003, 2004
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

/* === Header files. ====================================================== */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/poll.h>
#include <unistd.h>

#include "lib/device.h"
#include "lib/error.h"
#include "lib/misc.h"
#include "lib/queue.h"
#include "lib/vector.h"

/* === Type definitions, static variables, and static prototypes. ========= */

/* Used for associating ufd's with device's.  The entries in the device_devmap
 * list, have the same indices as their corresponding entries in device_ufds. */
typedef struct devmap {
    int readable:1;             /* If true, this fd is used for reading.    */
    int writable:1;             /* If true, this fd is used for writing.    */
    device_t* device;           /* Pointer to the device using this ufd.    */
} devmap_t;

static vector_t* device_list;   /* List of all known device's.  <device_t*> */
static vector_t* device_devmap; /* Map from ufd's to device's.  <devmap_t > */
static vector_t* device_ufds;   /* All monitored ufd's.         <ufd_t    > */

static int device_loop_done;    /* True if the event loop should exit.      */
static int device_inval_lists;  /* Have the device lists been edited?       */

/* Default readability handler. */
static int device_on_readable(device_t* d);

/* Default writability handler. */
static int device_on_writable(device_t* d);

/* Helper function for device_on_readable(). */
static int device_read(char* data, int count, va_list ap);

/* Helper function for device_on_writable(). */
static int device_write(char* data, int count, va_list ap);

/* Sort functions. */
static vector_sort_func_t device_sort_voidp;
static vector_sort_func_t device_sort_ufd;

/* Helper function for device_on_readable().  Interfaces queue_scan() with a
 * device's is_delim() function. */
static int device_is_delim(char data, va_list ap);

/* === Non-static function definitions. =================================== */

/* Initializes the device code. */
int device_init(void)
{
    int ret;

    try vector_new(&device_list,   0, sizeof(device_t*)); catch(error1);
    try vector_new(&device_devmap, 0, sizeof(devmap_t )); catch(error2);
    try vector_new(&device_ufds,   0, sizeof(ufd_t    )); catch(error3);

    vector_set_sort_func(device_list, device_sort_voidp);
    vector_set_sort_func(device_ufds, device_sort_ufd  );

    return E_OK;
error3: vector_del(device_devmap);
error2: vector_del(device_list  );
error1: return ret;
}

/* Cleans up after the device code. */
int device_cleanup(void)
{
    vector_del(device_list  );
    vector_del(device_devmap);
    vector_del(device_ufds  );
    return E_OK;
}

/* The main loop! */
int device_event_loop(void)
{
    int ret, poll_ret, i, count;
    devmap_t* devmap; ufd_t* ufds;

    /* Initialize counts and pointers.  Note that after inval_lists finishes,
     * control continues to next_round. */
    goto inval_lists;

    while (1) {
next_round:
        /* Determine which fd's need writability checking. */
        for (i = 0; i < count; ++i) {
            if (!devmap[i].writable) continue;
            ufds[i].events &= ~POLLOUT |
                (queue_count(devmap[i].device->write_buf) > 0 ? POLLOUT : 0);
        }

        /* Wait for any event.  If necessary, wait forever. */
        poll_ret = poll(ufds, count, -1);
        if (poll_ret <= 0) return -E_ERRNO;

        /* Process the event(s) that caused poll() to return. */
        for (i = 0; i < count; ++i) {
            device_t* devp = devmap[i].device;

            if (ufds[i].revents & POLLIN) {
                /* Call the readability handler. */
                ret = devp->on_readable != NULL ?  devp->on_readable(devp)
                                                : device_on_readable(devp);
                if (ret < E_OK) {
                    /* Handle error. */
                    fprintf(stderr, "\"%s\"->", devp->name);
                    error_print(ret, "on_readable()", NULL);
                    device_del(devp);
                }
                if (device_loop_done  ) goto loop_done;
                if (device_inval_lists) goto inval_lists;
                if (--poll_ret == 0   ) goto next_round;
            }

            if (ufds[i].revents & POLLOUT) {
                /* Call the writability handler. */
                ret = devp->on_writable != NULL ?  devp->on_writable(devp)
                                                : device_on_writable(devp);
                if (ret < E_OK) {
                    /* Handle error. */
                    fprintf(stderr, "\"%s\"->", devp->name);
                    error_print(ret, "on_writable()", NULL);
                    device_del(devp);
                }
                if (device_loop_done  ) goto loop_done;
                if (device_inval_lists) goto inval_lists;
                if (--poll_ret == 0   ) goto next_round;
            }

            if (ufds[i].revents & (POLLERR | POLLHUP | POLLNVAL)) {
                /* Error or disconnection. */
                error_print(-E_ERRNO, "poll()", "Error on device \"%s\".",
                            devp->name);
                device_del(devp);
                if (device_loop_done  ) goto loop_done;
                if (device_inval_lists) goto inval_lists;
                if (--poll_ret == 0   ) goto next_round;
            }
        }

        /* Control should never reach this point. */
        goto next_round;

inval_lists:
        /* Get up-to-date counts and pointers, and then start over. */
        count = vector_count(device_devmap);
        vector_index(device_ufds,   0, &ufds,   count);
        vector_index(device_devmap, 0, &devmap, count);
        goto next_round;

loop_done:
        /* No more device's, or some other reason to quit. */
        break;
    }
    return E_OK;
}

/* Creates a new device and stores it at *'d', assigns a name and path, and
 * inserts it into the list of device's. */
int device_new(device_t** d, char* name, char* path)
{
    int ret;
    device_t* dn;
    if (d == NULL || name == NULL || path == NULL) return -E_INVAL;

    /* Create a fresh device. */
    if ((dn = malloc(sizeof(*dn))) == NULL) {ret = -E_MEMORY; goto error1;}
    if ((dn->name = strdup(name))  == NULL) {ret = -E_MEMORY; goto error2;}
    if ((dn->path = strdup(path))  == NULL) {ret = -E_MEMORY; goto error3;}
    dn->on_readable    = NULL; dn->on_writable    = NULL;
    dn->on_cleanup_com = NULL; dn->on_cleanup_dev = NULL;
    dn->on_packet      = NULL; dn->is_delim       = NULL;
    dn->r_fd           =   -1; dn->w_fd           = -1;
    dn->com_state      = NULL; dn->dev_state      = NULL;
    try queue_new(&dn->read_buf,  0); catch(error4);
    try queue_new(&dn->write_buf, 0); catch(error5);

    /* Insert the device into device_list. */
    try vector_insert_sorted(device_list, &dn); catch(error6);
    device_inval_lists = 1;

    /* Yay. */
    *d = dn;
    return E_OK;
error6: queue_del(dn->write_buf);
error5: queue_del(dn->read_buf);
error4: free(dn->path);
error3: free(dn->name);
error2: free(dn);
error1: return ret;
}

/* Deallocates 'd', removes it from the event list, and, if defined, calls its
 * on_cleanup_*() functions. */
int device_del(device_t* d)
{
    int ret, pos;
    if (d == NULL) return -E_INVAL;

    /* Find and remove the device. */
    try vector_find_sorted(device_list, &d, &pos, NULL); catch(error1);
    if (ret == 0) return -E_SEARCH;
    try vector_discard(device_list, pos, 1); catch(error1);
    device_inval_lists = 1;

    /* Clean up the logical device state. */
    if (d->on_cleanup_dev != NULL) d->on_cleanup_dev(d);
    else if (d->dev_state != NULL) free(d->dev_state);

    /* Now do the hardware connection state. */
    if (d->on_cleanup_com != NULL) d->on_cleanup_com(d);
    else if (d->com_state != NULL) free(d->com_state);

    /* Remove this device from the ufd and devmap lists. */
    device_connect_fd(d, -1, -1);

    /* Deallocate memory used by this device. */
    queue_del(d->read_buf);
    queue_del(d->write_buf);
    free(d->name);
    free(d->path);
    free(d);

    return E_OK;
error1: return ret;
}

/* Connects 'd' to the named fd's.  If 'r_fd' equals 'w_fd', then the single fd
 * is used both for reading and writing.  If either is set to -1, then the
 * device is considered read-only or write-only.  If both are -1, then the
 * device is disconnected and disappears from the ufd lists. */
int device_connect_fd(device_t* d, int r_fd, int w_fd)
{
    ufd_t    ufd;
    devmap_t devmap;
    int ret, pos, step, count;
    if (d == NULL) return -E_INVAL;

    for (step = 1; step <= 4; ++step) {
        switch (step) {
        case 1:
            /* Step 1: remove old r_fd. */
            if (d->r_fd == -1) continue;        /* Was r_fd set? */
            ufd.fd = d->r_fd;
            break;
        case 2:
            /* Step 2: add new r_fd. */
            if (r_fd == -1) continue;           /* Is r_fd set? */
            ufd.fd = r_fd; ufd.events = POLLIN | POLLERR | POLLHUP;
            devmap.readable = 1;
            devmap.writable = (r_fd == w_fd);
            devmap.device   = d;
            break;
        case 3:
            /* Step 3: remove w_fd. */
            if (d->w_fd == -1)      continue;   /* Was w_fd set? */
            if (d->w_fd == d->r_fd) continue;   /* Was it separate from r_fd? */
            ufd.fd = d->w_fd;
            break;
        case 4:
            /* Step 4: add new w_fd. */
            if (w_fd == -1)   continue;         /* Is w_fd set? */
            if (w_fd == r_fd) continue;         /* Is it separate from r_fd? */
            ufd.fd = w_fd; ufd.events = POLLOUT | POLLERR | POLLHUP;
            devmap.readable = 0;
            devmap.writable = 1;
            devmap.device   = d;
        }

        /* Find the fd entry in device_ufds and device_devmap. */
        ret = vector_find_sorted(device_ufds, &ufd, &pos, NULL);
        if (ret < E_OK) return ret;

        switch (step) {
        case 1: case 3:
            /* Remove it. */
            if (ret != 1) return -E_SEARCH;
            vector_discard(device_ufds,   pos, 1);
            vector_discard(device_devmap, pos, 1);
            break;
        case 2: case 4:
            /* Add it. */
            if (ret != 0) return -E_SEARCH;
            vector_insert(device_ufds,   pos, &ufd,    1);
            vector_insert(device_devmap, pos, &devmap, 1);
        }
    }

    /* Record the changes. */
    d->r_fd = r_fd;
    d->w_fd = w_fd;

    device_inval_lists = 1;
    return E_OK;
}

/* === Static function definitions. ======================================= */

/* Default readability handler. */
static int device_on_readable(device_t* d)
{
    char* data;
    int ret, pos;
    if (d == NULL) return -E_INVAL;

    /* Read data from the device. */
    try queue_fill(d->read_buf, device_read, d); catch(error1);
    pos = -ret;

    while (1) {
        /* Find the end of packet. */
        ret = queue_scan(d->read_buf, pos, device_is_delim, d, d->is_delim);
        if (ret == -E_SEARCH) break;
        if (ret < E_OK)       return ret;
        
        /* End of packet found. */
        try queue_index(d->read_buf, &data, ret + 1); catch(error1);
        try d->on_packet(d, data, ret + 1);           catch(error1);
        try queue_discard(d->read_buf, ret + 1);      catch(error1);
        pos = 0;
    }

    return E_OK;
error1: return ret;
}

/* Default writability handler. */
static int device_on_writable(device_t* d)
{
    return queue_drain(d->write_buf, device_write, d);
}

/* Helper function for device_on_readable(). */
static int device_read(char* data, int count, va_list ap)
{
    int ret;
    device_t* d = va_arg(ap, device_t*);
    if (d == NULL || d->r_fd < 0) return -E_INVAL;

    ret = read(d->r_fd, data, count);
    if (ret < 0)  return -E_ERRNO;      /* Read failure. */
    if (ret == 0) return -E_USER + 0;   /* End of file.  */
                  return ret;           /* Success.      */
}

/* Helper function for device_on_writable(). */
static int device_write(char* data, int count, va_list ap)
{
    int ret;
    device_t* d = va_arg(ap, device_t*);
    if (d == NULL || d->w_fd < 0) return -E_INVAL;

    ret = write(d->w_fd, data, count);
    if (ret < 0)  return -E_ERRNO;      /* Read failure. */
    if (ret == 0) return -E_USER + 0;   /* End of file.  */
                  return ret;           /* Success.      */
}

/* Sorts pointers in ascending order. */
static int _device_sort_voidp(const char* a, const char* b)
{
    return SGN(a - b);
}
static vector_sort_func_t device_sort_voidp =
      (vector_sort_func_t)_device_sort_voidp;
/* This cast is necessary since it's uncouth to do pointer math with void*'s. */

/* Sorts ufd's by file descriptor. */
static int _device_sort_ufd(const ufd_t* a, const ufd_t* b)
{
    return SGN(a->fd - b->fd);
}
static vector_sort_func_t device_sort_ufd =
      (vector_sort_func_t)_device_sort_ufd;

/* Helper function for device_on_readable().  Interfaces queue_scan() with a
 * device's is_delim() function. */
static int device_is_delim(char data, va_list ap)
{
    device_t* d              = va_arg(ap, typeof(d   ));
    typeof(d->is_delim) func = va_arg(ap, typeof(func));
    return func(d, data);
}

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
