/* lib/device.h
 *
 * Device management and input/output abstraction.
 *
 * Copyright (C) 2003, 2004
 * Zorica Majdov <zmajdov@virtuoso.com>
 * Andy Goth     <unununium@openverse.com>
 *
 * This code is available under the GNU General Public License; see COPYING. */

#ifndef SEEN_LIB_DEVICE_H
#define SEEN_LIB_DEVICE_H

#include <unistd.h>

#include "lib/queue.h"

/* "ufd_t" is shorter to type than "struct pollfd".  :^)  The name derives
 * from the man page for poll(): it lists the first argument as "ufds". */
typedef struct pollfd ufd_t;

/* This struct interfaces the various data sources/sinks with the main event
 * loop, which multiplexes all input, output, and timer events using the poll()
 * call.  It provides buffering and packetizing mechanisms for the default
 * handlers to use.
 *
 * dev_state is used by the protocol decoder module, and com_state is for the
 * hardware interface code.  r_fd and w_fd are passed to system calls to
 * actually talk to the physical device.
 *
 * I hear you ask, what's the difference between protocol decoding and hardware
 * interfacing?  Okay, I don't really hear you, but at any rate this is the
 * question you should ask.  Protocol decoding code has some knowledge of how
 * the device serializes data; it's responsible for converting the bitstream
 * into useful values.  Hardware interfacing code doesn't know squat about the
 * device, just that it's connected via the serial port, USB, TCP/IP, CPIP, or
 * what-not; it's responsible for providing the bitstream. */
struct device; typedef struct device device_t;
struct device {
    /* Name used in displayed messages. */
    char* name;

    /* Path to device node, if applicable. */
    char* path;

    /* If non-NULL, overrides the default handler for POLLIN. */
    int (*on_readable)(device_t* d);

    /* If non-NULL, overrides the default handler for POLLOUT. */
    int (*on_writable)(device_t* d);

    /* If non-NULL, overrides the default handler for destroying this device.
     * on_cleanup_dev() is for cleaning up dev_state, and on_cleanup_com() is
     * for cleaning up com_state. */
    int (*on_cleanup_dev)(device_t* d);
    int (*on_cleanup_com)(device_t* d);

    /* Called by the default on_readable() after each packet is received.
     * 'data' points to the start of the packet buffer, and 'len' gives the
     * number of bytes in the packet. */
    int (*on_packet)(device_t* d, char* data, int len);

    /* Called by the default on_readable() handler to determine if a given byte
     * is the end of a packet (the delimiter).  If this function returns 1,
     * then everything received since the end of the last packet up to the
     * given byte is passed to on_packet() for processing. */
    int (*is_delim)(device_t* d, char data);

    /* Used by the default on_readable() handler to store packets as they
     * are in the process of being received and processed. */
    queue_t* read_buf;

    /* Used by the default on_writable() handler to store data while it is
     * waiting to be written to the device. */
    queue_t* write_buf;
    
    /* For poll(), read(), write(), and other system calls.  r_fd is used for
     * reading, and w_fd is for writing.  For devices that use a single read-
     * write fd, r_fd equals w_fd.  r_fd and w_fd can be set to -1 to indicate
     * that the device is read-only, write-only, or not (yet) connected. */
    int r_fd, w_fd;

    /* Logical state information regarding the device protocol; used by
     * on_packet(), is_packet_end(), and on_cleanup_dev(). */
    void* dev_state;

    /* Information about the connection to the hardware; used by on_readable(),
     * on_writable(), and on_cleanup_com().  The name is derived from the
     * colloquial DOSism "COM port".  */
    void* com_state;
};

/* Initializes the device code. */
extern int device_init(void);

/* Cleans up after the device code. */
extern int device_cleanup(void);

/* The main loop! */
extern int device_event_loop(void);

/* Creates a new device and stores it at *'d', assigns a name and path, and
 * inserts it into the list of device's. */
extern int device_new(device_t** d, char* name, char* path);

/* Deallocates 'd', removes it from the event list, and, if defined, calls its
 * on_cleanup_*() functions. */
extern int device_del(device_t* d);

/* Connects 'd' to the named fd's.  If 'r_fd' equals 'w_fd', then the single fd
 * is used both for reading and writing.  If either is set to -1, then the
 * device is considered read-only or write-only.  If both are -1, then the
 * device is disconnected and disappears from the ufd lists. */
extern int device_connect_fd(device_t* d, int r_fd, int w_fd);

#endif

/* ======================================================================== */

/* vim: set ts=4 sts=4 sw=4 tw=80 et: */
/* EOF */
